(({ theme, Toolshed, t }) => {
  /**
   * Function to create an HTMLElement to serve as the expand/collapse toggle
   * for an AccordionItem.
   *
   * @return {HTMLElement}
   *   An HTMLElement to serve as the expand/collapse toggle for an
   *   AccordionItem.
   */
  theme.accordionToggler = () => {
    // Create top-level button element.
    const btn = document.createElement('BUTTON');
    btn.classList.add('accordion-item__toggler');

    // Create inner content to expand accordion item.
    const expand = document.createElement('SPAN');
    expand.classList.add('accordion-item__toggler-inner');
    expand.classList.add('accordion-item__toggler-inner--expand');
    expand.textContent = t('expand');
    btn.appendChild(expand);

    // Create inner content to collapse accordion item.
    const collapsed = document.createElement('SPAN');
    collapsed.classList.add('accordion-item__toggler-inner');
    collapsed.classList.add('accordion-item__toggler-inner--collapse');
    collapsed.textContent = t('collapse');
    btn.appendChild(collapsed);

    return btn;
  };

  /**
   * Defines a collapsible accordion item.
   */
  class AccordionItem {
    /**
     * Class representing a single item of an Accordion which is expandable and
     * collapsible. This item maintains its components and the state of the
     * open and close states.
     *
     * @param {Element} el
     *   The whole accordion item.
     * @param {Element} expander
     *   Element to use for toggling the body content.
     * @param {Element} body
     *   The content area of the accordion item which is revealed or hidden
     *   when the accordion item is expanded or collapsed.
     * @param {Accordion|AccordionItem} parent
     *   Parent accordion instance.
     * @param {Object} settings
     *   Object of configurations to use to control the behavior of the
     *   accordion item.
     */
    constructor(el, expander, body, parent, settings) {
      this.isExpanded = true;

      this.el = el;
      this.body = body;
      this.expander = expander;
      this.parent = parent;
      this.useAnimation = settings.animate || true;
      this.transitionDuration = settings.transitionDuration || '0.3s';
      this.children = [];

      this.el.classList.add('accordion-item');
      this.el.classList.add('accordion-item--collapsible');

      if (this.useAnimation) {
        this._height = this.body.clientHeight;
      }

      this.body.classList.add('accordion-item__body');

      this.expander.ariaExpanded = this.isExpanded.toString();
      this.onToggleExpand = AccordionItem[settings.exclusive ? 'onClickExclusive' : 'onClickToggle'].bind(this);

      expander.addEventListener('click', this.onToggleExpand);
      this.body.addEventListener('transitionend', AccordionItem.onTransitionEnd);
    }

    /**
     * Add a child accordion item.
     *
     * @param {AccordionItem} child
     *   Child accordion item to add.
     */
    addChild(child) {
      this.children.push(child);
    }

    /**
     * Expand the accordion item's content display.
     *
     * @param {bool} allowAnimate
     *   Animate expand animations if an animation is configured.
     */
    expand(allowAnimate = true) {
      this.el.classList.add('accordion-item--expanded');
      this.el.classList.remove('accordion-item--collapsed');
      this.expander.ariaExpanded = 'true';
      this.isExpanded = true;
      this.expander.classList.add('accordion-item__toggler--expanded');
      this.expander.classList.remove('accordion-item__toggler--collapsed');

      if (allowAnimate && this.useAnimation) {
        Object.assign(this.body.style, {
          display: null,
          boxSizing: 'border-box',
          height: '0px',
          paddingTop: '0px',
          paddingBottom: '0px',
          transition: `height ${this.transitionDuration}, padding ${this.transitionDuration}`,
          overflow: 'hidden',
        });

        requestAnimationFrame(() => {
          requestAnimationFrame(() => {
            Object.assign(this.body.style, {
              height: `${this._height || 50}px`,
              paddingTop: null,
              paddingBottom: null,
            });
          });
        });
      }
      else {
        this.body.style.display = '';
      }
    }

    /**
     * Collapse the accordion item's content display.
     *
     * @param {bool} allowAnimate
     *   Should the collapse functionality expect to animate the transitions
     *   if the accordion item is setup to do animations.
     */
    collapse(allowAnimate = true) {
      this.el.classList.add('accordion-item--collapsed');
      this.el.classList.remove('accordion-item--expanded');
      this.expander.ariaExpanded = 'false';
      this.isExpanded = false;
      this.expander.classList.add('accordion-item__toggler--collapsed');
      this.expander.classList.remove('accordion-item__toggler--expanded');

      if (allowAnimate && this.useAnimation) {
        this._height = this.body.clientHeight;

        // Browsers are not able to perform transitions on elements with
        // auto-height, because the browser doesn't have a value to transition
        // between. Set the current height, wait for the an animation frame for
        // the change to take effect and then set the transition.
        Object.assign(this.body.style, {
          height: `${this._height}px`,
          boxSizing: 'border-box',
          transition: 'none',
        });

        requestAnimationFrame(() => {
          Object.assign(this.body.style, {
            transition: `height ${this.transitionDuration}, padding ${this.transitionDuration}`,
          });

          requestAnimationFrame(() => {
            Object.assign(this.body.style, {
              height: '0px',
              paddingTop: '0px',
              paddingBottom: '0px',
              overflow: 'hidden',
            });
          });
        });
      }
      else {
        this.body.style.display = 'none';
      }

      // Recursively close all the open child accordions.
      this.children.forEach((child) => {
        if (child.isExpanded) child.collapse(false);
      });
    }

    /**
     * Expand and collapse of accordion items, without concern to the state of
     * other accordion items. Multiple items can be open at once.
     *
     * @param {ClickEvent} e
     *   DOM click event object, containing information about the item being
     *   being clicked with the collapse functionality.
     */
    static onClickToggle(e) {
      e.preventDefault();

      if (this.isExpanded) {
        this.collapse();
      }
      else {
        this.expand();
      }
    }

    /**
     * Expand and collapse of accordion items but ensure only one item is open
     * at a time.
     *
     * @param {ClickEvent} e
     *   DOM click event object, containing information about the item being
     *   being clicked with the collapse functionality.
     */
    static onClickExclusive(e) {
      e.preventDefault();

      if (this.isExpanded) {
        this.collapse();
      }
      else {
        // Recursively close all the open child accordions.
        this.parent.children.forEach((child) => {
          if (child !== this && child.isExpanded) child.collapse();
        });

        this.expand();
      }
    }

    /**
     * Element has completed its CSS transition, perform final callback.
     *
     * @param {TransitionEvent} e
     *   The event object with the transition information.
     */
    static onTransitionEnd(e) {
      if (e.propertyName === 'height') {
        const { style } = e.target;

        if (style.height === '0px') {
          style.display = 'none';
        }
        else {
          style.display = '';
          style.boxSizing = '';
          style.height = '';
        }

        style.overflow = '';
        style.transition = '';
      }
    }

    /**
     * Teardown any events and alterations this accordion item made to the DOM.
     */
    destroy() {
      this.el.classList.remove('accordion-item');
      this.el.classList.remove('accordion-item--collapsed');
      this.el.classList.remove('accordion-item--expanded');
      this.el.classList.remove('accordion-item--collapsible');
      this.expander.removeAttribute('aria-expanded');
      this.expander.classList.remove('accordion-item__toggler--expanded');
      this.expander.classList.remove('accordion-item__toggler--collapsed');
      // TODO: remove expander from DOM if it was added to the DOM by the
      // AccordionItem. Will require some refactoring, since currently the
      // expander is added outside of AccordionItem and then passed in, so
      // currently AccordionItem is not aware of whether this should be removed
      // when destroying itself.

      Object.assign(this.body.style, {
        display: '',
        height: '',
        padding: '',
        overflow: '',
      });

      this.expander.removeEventListener('click', this.onToggleExpand);
      this.body.removeEventListener('transitionend', AccordionItem.onTransitionEnd);
    }
  }

  /**
   * Defines a non-collapsible accordion item.
   */
  class AccordionItemNonCollapsible {
    /**
     * Class representing a single item of an Accordion which is not expandable
     * or collapsible.
     *
     * @param {Element} el
     *   The whole accordion item.
     * @param {Accordion|AccordionItem} parent
     *   Parent accordion instance.
     */
    constructor(el, parent) {
      this.el = el;
      this.parent = parent;
      this.isExpanded = false;
      this.children = [];

      this.el.classList.add('accordion-item');
      this.el.classList.add('accordion-item--non-collapsible');
    }

    /**
     * Add a child accordion item.
     *
     * @param {AccordionItem} child
     *   Child accordion item to add.
     */
    addChild(child) {
      this.isExpanded = true;
      this.children.push(child);
    }

    /**
     * This class implements an expand function in order to have the same
     * interface as the AccordionItem class, but since this is a
     * non-collapsible accordion item, it does nothing.
     */
    expand() {} // eslint-disable-line class-methods-use-this

    /**
     * This class implements a collapse function in order to have the same
     * interface as the AccordionItem class, but since this is a
     * non-collapsible accordion item, it just collapses any children.
     */
    collapse() {
      // Recursively close all the open child accordions.
      this.children.forEach((child) => {
        if (child.isExpanded) child.collapse(false);
      });
    }

    /**
     * Teardown any alterations this accordion item made to the DOM.
     */
    destroy() {
      this.el.classList.remove('accordion-item');
      this.el.classList.remove('accordion-item--non-collapsible');
    }
  }

  /**
   * Accordion object definition.
   */
  Toolshed.Accordion = class Accordion {
    /**
     * Create an accordion of expandable and collapsible items.
     *
     * @param {Element} el
     *   DOM element to turn into an accordion.
     * @param {Object} configs
     *   A configuration options which determine the options:
     *   {
     *     initOpen: {bool|selector} [true],
     *       // Should accordion panels start of initially open.
     *     exclusive: {bool} [true],
     *       // Accordion only has one item open at a time.
     *     animate: {bool} [true],
     *       // Should the accordions expand and collapse animate.
     *     transitionDuration: {string} ['0.3s']
     *       // How long the expand/collapse transitions should last. Should be
     *       // a duration that a CSS transition can accept (such as '300ms' or
     *       // '0.3s').
     *     expandOpt: {DOMString|function|object}
     *       // The selector to find the element to toggle the collapse and expand, OR
     *       // A callback function to create the toggler element, OR
     *       // An object with `create` and `attacher` functions to create and
     *       // place the expander into the DOM.
     *     itemSelector: {DOMString}
     *       // The selector to use when locating each of the accordion items.
     *     bodySelector: {DOMString}
     *       // The selector to use to find the item body within the item context.
     *   }
     */
    constructor(el, configs) {
      this.el = el;
      this.items = new Map();
      this.children = [];

      this.configs = {
        exclusive: true,
        initOpen: true,
        animate: true,
        transitionDuration: '0.3s',
        ...configs,
      };

      const {
        exclusive,
        initOpen,
        itemSelector,
        bodySelector,
      } = this.configs;

      this.el.classList.add('accordion');

      let expandOpt = this.configs.expandOpt || Drupal.theme.accordionToggler;

      // If the expand option is just a selector, then the element is expected
      // to already be in the document (otherwise not findable).
      if (typeof expandOpt === 'string' || expandOpt instanceof String) {
        this.getExpander = (item) => item.querySelector(expandOpt);
      }
      else {
        if (typeof expandOpt === 'function') {
          expandOpt = { create: expandOpt };
        }

        expandOpt.attacher = expandOpt.attacher || ((expander, item) => {
          item.insertBefore(expander, item.querySelector(bodySelector));
        });

        this.getExpander = (item) => {
          let expander = expandOpt.create(item);

          // If the result of the create callback, is not a HTMLElement,
          // create an HTML element from the result (assumes a string of HTML).
          if (!(expander instanceof HTMLElement)) {
            const temp = document.createElement('DIV');
            temp.innerHTML = String(expander);
            expander = temp.firstElementChild;
            temp.removeChild(expander);
          }

          // If expander is not already in the DOM, use the attacher callback
          // to insert it.
          if (expander && !expander.parentNode) {
            expandOpt.attacher(expander, item);
          }

          return expander;
        };
      }

      // Build the accordion items allowing for a nested tree structure.
      Array.prototype.forEach.call(el.querySelectorAll(itemSelector), this.buildTreeItem, this);

      if (!initOpen || exclusive) {
        this.children.forEach((child, i) => {
          if (!initOpen || i !== 0) child.collapse(false);
        });
      }
    }

    /**
     * Add a child accordion item.
     *
     * @param {AccordionItem} child
     *   Child accordion item to add.
     */
    addChild(child) {
      this.children.push(child);
    }

    /**
     * Find the parent accordion item for this element.
     *
     * @param {Element} el
     *   The element to search for the closest matching selector for.
     * @param {DOMSelector} selector
     *   Selector to use when identifying an accordion item.
     *
     * @return {Element|null}
     *   If an appropriate parent matching the item selector is found within
     *   the confines of the accordion element, it is returned. Otherwise,
     *   return the Element wrapping the accordion.
     */
    findParent(el) {
      let cur = el;
      const { itemSelector } = this.configs;

      do {
        cur = cur.parentElement;
      } while (cur && cur !== this.el && !cur.matches(itemSelector));

      return cur;
    }

    /**
     * Recursively add accordion items to the tree structure.
     *
     * @param {Element} item
     *   The accordion item wrapper transform into an accordion item.
     *
     * @return {AccordionItem|AccordionItemNonCollapsible}
     *   The current accordion item.
     */
    buildTreeItem(item) {
      let accord = this.items.get(item);

      if (accord) {
        return accord;
      }

      const { itemSelector, bodySelector } = this.configs;
      const parent = this.findParent(item, itemSelector);

      // Only create an accordion item if it has a parent item which exists
      // within the Accordion scope.
      if (!parent) throw new Error('Unable to find a parent for the item in buildTreeItem.');

      const itemParent = (parent === this.el)
        ? this : this.items.get(parent) || this.buildTreeItem(parent);

      if (!itemParent) throw new Error('No valid accordion parent available, and not a descendent of the accordion element.');

      // Only create a collapsible accordion item if it has a has a body to
      // expand, and an expander (expand/collapse toggle). Otherwise create a
      // non-collapsible accordion item.
      const body = item.querySelector(bodySelector);

      if (body) {
        const expander = this.getExpander(item);

        if (expander) {
          // This accordion item should be made collapsible.
          accord = new AccordionItem(item, expander, body, itemParent, {
            animate: this.configs.animate,
            transitionDuration: this.configs.transitionDuration,
            exclusive: this.configs.exclusive,
          });
        }
      }

      if (!accord) {
        // This accordion item should be made non-collapsible.
        accord = new AccordionItemNonCollapsible(item, itemParent);
      }

      // Add the accordion item to the accordion's 'items' list, and to its
      // parent's list of children.
      this.items.set(item, accord);
      itemParent.addChild(accord);

      return accord;
    }

    /**
     * Teardown any changes to the DOM as well as clearing AccordionItems.
     */
    destroy() {
      this.el.classList.remove('accordion');
      this.items.forEach((item) => item.destroy());
    }
  };
})(Drupal);
