<?php

namespace Drupal\toolshed_media\Plugin\Field\FieldFormatter;

use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Component\Utility\Html;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Field\Plugin\Field\FieldFormatter\EntityReferenceFormatterBase;
use Drupal\toolshed_media\Utility\FileHelper;

/**
 * A Field formatter for displaying media or image files with more options.
 *
 * @FieldFormatter(
 *   id = "toolshed_image_style",
 *   label = @Translation("Media image style"),
 *   field_types = {
 *     "entity_reference",
 *   },
 * )
 */
class MediaImageStyleFormatter extends EntityReferenceFormatterBase implements ContainerFactoryPluginInterface {

  /**
   * Service which retrieves and maintains Drupal system configurations.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Service which manages various entity types and their handlers.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Service for managing the various entity view modes.
   *
   * @var \Drupal\Core\Entity\EntityDisplayRepository
   */
  protected $entityDisplayRepo;

  /**
   * Media image style formatter object.
   *
   * @param array $configuration
   *   Array of plugin configuration options which includes the field formatter
   *   definition information:
   *     - $field_definition
   *     - $settings
   *     - $label
   *     - $view_mode
   *     - $third_party_settings
   *   These are the usual values for constructing a standard field formatter.
   * @param string $plugin_id
   *   The plugin_id for the formatter.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Service for retrieving and maintain Drupal system configurations.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity manager service, for getting entity definitions and handlers.
   * @param \Drupal\Core\Entity\EntityDisplayRepositoryInterface $entity_display_repository
   *   Manage entity view mode configurations and displays.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, ConfigFactoryInterface $config_factory, EntityTypeManagerInterface $entity_type_manager, EntityDisplayRepositoryInterface $entity_display_repository) {
    parent::__construct(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings']
    );

    $this->configFactory = $config_factory;
    $this->entityTypeManager = $entity_type_manager;
    $this->entityDisplayRepo = $entity_display_repository;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('entity_display.repository')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $fieldDef) {
    $fieldStorageDef = $fieldDef->getFieldStorageDefinition();
    $settings = $fieldDef->getSetting('handler_settings');

    return $fieldStorageDef->getSettings()['target_type'] === 'media'
      && (empty($settings['target_bundles']) || in_array('image', $settings['target_bundles']));
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $entityType = $this->fieldDefinition->getSetting('target_type');
    $elements = [];

    if ($entityType !== 'media') {
      return [];
    }

    $styleName = $this->getSetting('image_style');
    $viewBuilder = $this->entityTypeManager->getHandler('media', 'view_builder');

    foreach ($this->getEntitiesToView($items, $langcode) as $delta => $entity) {
      if ($entity->getSource()->getPluginId() === 'image') {
        if ($fileHelper = FileHelper::fromEntity($entity)) {
          $fileData = $fileHelper->getData();

          $elements[$delta] = [
            '#theme' => 'image',
            '#uri' => $fileHelper->getUri(),
          ];

          foreach (['alt', 'title', 'width', 'height', 'attributes'] as $attrName) {
            if (!empty($fileData[$attrName])) {
              $elements[$delta]["#$attrName"] = $fileData[$attrName];
            }
          }

          if (!empty($styleName) && $styleName !== 'none') {
            $elements[$delta]['#theme'] = 'image_style';
            $elements[$delta]['#style_name'] = $styleName;
          }
        }
      }

      // Fallback, to use the media view builder. Used if media entity is not
      // targeting an image or if we are unable to load the needed file info.
      if (!isset($elements[$delta])) {
        $elements[$delta] = $viewBuilder->view($entity, $this->getSetting('fallback_display'));
      }
    }

    return $elements;
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'image_style' => '',
      'fallback_display' => 'default',
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $styleId = $this->getSetting('image_style');

    if (!empty($styleId)) {
      $style = $this->entityTypeManager
        ->getStorage('image_style')
        ->load($styleId);
    }

    $styleLabel = $style ? $style->label() : $this->t('Original');

    return [
      $this->t('Image style:') . ' ' . $styleLabel,
      $this->t('Fallback display: %fallback_display', ['%fallback_display' => $this->getSetting('fallback_display')]),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $form = parent::settingsForm($form, $form_state);
    $fieldStorageDef = $this->fieldDefinition->getFieldStorageDefinition();
    $styleStorage = $this->entityTypeManager->getStorage('image_style');

    $styleOpts = ['none' => $this->t('Original')];
    foreach ($styleStorage->loadMultiple() as $style) {
      $styleOpts[$style->id()] = $style->label();
    }

    // If being used with embed styles, filter the allowed embed styles to
    // only the allowed styles. This prevents the usage of image presets
    // that don't make sense here or, could be dangerous in an embed situation.
    if ($this->viewMode === '_entity_embed') {
      $embedStyles = $this->configFactory->get('toolshed.media.embed_styles');
      $filterStyles = $embedStyles->get('styles');

      if (!empty($filterStyles)) {
        $styleOpts = $embedStyles->get('filter_method') === 'only_listed'
          ? array_intersect_key($styleOpts, $filterStyles)
          : array_diff_key($styleOpts, $filterStyles);
      }
    }

    $form['image_style'] = [
      '#type' => 'select',
      '#title' => $this->t('Image style'),
      '#options' => $styleOpts,
      '#default_value' => $this->getSetting('image_style'),
    ];

    $targetType = $fieldStorageDef->getSettings()['target_type'];
    $viewModes = ['default' => $this->t('Default')];
    foreach ($this->entityDisplayRepo->getViewModes($targetType) as $viewModeId => $viewMode) {
      $viewModes[$viewModeId] = Html::escape($viewMode['label']);
    }

    $form['fallback_display'] = [
      '#type' => 'select',
      '#title' => $this->t('Fallback display mode'),
      '#options' => $viewModes,
      '#default_value' => $this->getSetting('fallback_display'),
      '#description' => $this->t('If the the media entity is not an image, image styles cannot be applied and will use this view mode to render as a fallback.'),
    ];

    return $form;
  }

}
