<?php

namespace Drupal\toolshed_search\Form;

use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Component\Render\FormattableMarkup;
use Drupal\toolshed_search\Plugin\Block\ToolshedSearchBlock;

/**
 * Form class for creating a keyword search from an exposed view filter.
 */
class ToolshedSearchFilterForm extends FormBase {

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * Create a new instance of the ToolshedSearchFilterForm.
   *
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack which include the current request.
   */
  public function __construct(RequestStack $request_stack) {
    $this->request = $request_stack->getCurrentRequest();
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('request_stack')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'toolshed_search_filter_form';
  }

  /**
   * Disable the form build tokens as they do not apply for this keyword form.
   *
   * @param array $form
   *   Structure of the form after it has been processed.
   *
   * @return array
   *   Altered form to render. Should be the form without any of the build
   *   and token information typically used, since it is only to populate the
   *   keyword search.
   */
  public static function afterBuild(array $form) {
    // Form using POST HTTP method, these items are not used and just end up
    // cluttering the URL query response.
    $form['form_id']['#access'] = FALSE;
    $form['form_build_id']['#access'] = FALSE;
    $form['form_token']['#access'] = FALSE;
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, ToolshedSearchBlock $block = NULL) {
    $config = $block->getConfiguration();
    $view = $block->getView();

    // Setup the form meta data.
    $form['filters'] = [];
    $form['#token'] = FALSE;
    $form['#after_build'][] = static::class . '::afterBuild';

    // Determine the URL where the from should submit to.
    $form['#action'] = $config['action'] === 'current_page'
      ? $this->request->getPathInfo()
      : $view->getUrl()->toString();

    // Create the filter elements for the form.
    foreach ($block->getFilters() as $name => $filter) {
      $filter->buildExposedForm($form['filters'], $form_state);

      if ($info = $filter->exposedInfo()) {
        $form['#info']["filter-{$name}"] = $info;
      }
    }

    // Setup the search filter form submit button.
    $buttonClasses = $config['submit_class_names'] ?? [];
    $buttonClasses[] = 'button';
    $buttonClasses[] = 'button--search';

    $form['actions']['search'] = [
      '#type' => 'html_button',
      '#attributes' => [
        'type' => 'submit',
        'class' => $buttonClasses,
      ],

      'content' => [
        '#type' => 'html_tag',
        '#tag' => 'span',
        '#attributes' => [
          'class' => ['button__text'],
        ],
        '#plain_text' => $config['submit_label'],
      ],
    ];

    // Apply variant information to the form so theme suggestion and preprocess
    // can utilize the variant data.
    if (!empty($config['variant'])) {
      $form['#block_variant'] = $config['variant'];
      $form['actions']['search']['#attributes']['aria-label'] = new FormattableMarkup('@variant @label', [
        '@label' => $config['submit_label'],
        '@variant' => $config['variant'],
      ]);
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Form does uses GET (and doesn't use caching), so will not be processed.
    // This means this method will never be called, and the actual processing
    // will happen on the destination page.
  }

}
